
import time
import json
import os
from os.path import basename, getsize, join, dirname, isfile, getctime
from pathlib import Path
from random import random
import requests

from geoseeq.knex import GeoseeqGeneralError
from geoseeq.constants import FIVE_MB
from geoseeq.utils import md5_checksum
from concurrent.futures import ThreadPoolExecutor, as_completed
from .utils import *
from geoseeq.file_system_cache import GEOSEEQ_CACHE_DIR
from .file_chunker import FileChunker


class ResumableUploadTracker:

    def __init__(self, filepath, chunk_size, upload_target_uuid, tracker_file_prefix="gs_resumable_upload_tracker"):
        self.open, self.upload_started = True, False
        self.upload_id, self.urls, self.is_atomic_upload = None, None, None
        self.upload_target_uuid = upload_target_uuid
        self.filepath = filepath
        self.tracker_file_dir = join(GEOSEEQ_CACHE_DIR, 'upload')
        self.tracker_file = join(
            self.tracker_file_dir,
            tracker_file_prefix + f".{upload_target_uuid}.{chunk_size}.{getsize(filepath)}." + basename(filepath)
        )
        try:
            os.makedirs(self.tracker_file_dir, exist_ok=True)
        except Exception as e:
            logger.warning(f'Could not create resumable upload tracker directory. {e}')
            self.open = False
        self._loaded_parts = {}
        self._load_parts_from_file()

    def start_upload(self, upload_id, urls, is_atomic_upload=False):
        if not self.open:
            return
        if self.upload_started:
            raise GeoseeqGeneralError("Upload has already started.")
        self.upload_started = True
        blob = dict(upload_id=upload_id,
                    urls=urls,
                    is_atomic_upload=is_atomic_upload,
                    upload_target_uuid=self.upload_target_uuid,
                    start_time=time.time())
        serialized = json.dumps(blob)
        with open(self.tracker_file, "w") as f:
            f.write(serialized + "\n")
        self.upload_id, self.urls, self.is_atomic_upload = upload_id, urls, is_atomic_upload

    def add_part(self, part_upload_info):
        if not self.open:
            return
        part_id = part_upload_info["PartNumber"]
        serialized = json.dumps(part_upload_info)
        with open(self.tracker_file, "a") as f:
            f.write(serialized + "\n")
        self._loaded_parts[part_id] = part_upload_info
        if len(self._loaded_parts) == len(self.urls):
            self.cleanup()
            self.open = False
    
    def _load_parts_from_file(self):
        if not isfile(self.tracker_file):
            return
        with open(self.tracker_file, "r") as f:
            header_blob = json.loads(f.readline())
            self.upload_id, self.urls, self.is_atomic_upload = (
                header_blob["upload_id"], header_blob["urls"], header_blob["is_atomic_upload"]
            )
            start_time = header_blob["start_time"]
            if (time.time() - start_time) > (60 * 60 * 23):
                logger.warning(f"Tracker file {self.tracker_file} is too old. Deleting.")
                os.remove(self.tracker_file)
                return
            self.upload_started = True
            for line in f:
                blob = json.loads(line)
                part_id = blob["PartNumber"]
                self._loaded_parts[part_id] = blob
    
    def part_has_been_uploaded(self, part_number):
        if not self.open:
            return False
        return part_number in self._loaded_parts
    
    def get_part_info(self, part_number):
        return self._loaded_parts[part_number]
    
    def cleanup(self):
        if not self.open:
            return
        try:
            os.remove(self.tracker_file)
        except FileNotFoundError:
            pass




import time
import json
import os
from os.path import basename, getsize, join, dirname, isfile, getctime
from pathlib import Path
from random import random
import requests

from geoseeq.knex import GeoseeqGeneralError
from geoseeq.constants import FIVE_MB
from geoseeq.utils import md5_checksum
from concurrent.futures import ThreadPoolExecutor, as_completed
from .utils import *
from geoseeq.file_system_cache import GEOSEEQ_CACHE_DIR
from .file_chunker import FileChunker



class ResumableDownloadTracker:

    def __init__(self, chunk_size, download_target_id, target_local_path, tracker_file_prefix="gs_resumable_download_tracker"):
        self.open, self.download_started = True, False
        self.download_target_id = download_target_id
        self.target_local_path = target_local_path
        self.tracker_file_dir = join(GEOSEEQ_CACHE_DIR, 'download')
        self.tracker_file = join(
            self.tracker_file_dir,
            tracker_file_prefix + f".{download_target_id}.{chunk_size}." + basename(target_local_path)
        )
        try:
            os.makedirs(self.tracker_file_dir, exist_ok=True)
        except Exception as e:
            logger.warning(f'Could not create resumable download tracker directory. {e}')
            self.open = False
        self._loaded_parts = {}
        self._load_parts_from_file()

    def start_download(self, download_url):
        if not self.open:
            return
        if self.download_started:
            raise GeoseeqGeneralError("Download has already started.")
        self.download_started = True
        blob = dict(download_url=download_url,
                    download_target_id=self.download_target_id,
                    start_time=time.time())
        serialized = json.dumps(blob)
        with open(self.tracker_file, "w") as f:
            f.write(serialized + "\n")
        self.download_url = download_url
        return self
    
    def add_part(self, part_download_info):
        if not self.open:
            assert False, "Cannot add part to closed ResumableDownloadTracker"
        part_id = part_download_info["part_number"]
        serialized = json.dumps(part_download_info)
        with open(self.tracker_file, "a") as f:
            f.write(serialized + "\n")
        self._loaded_parts[part_id] = part_download_info

    def _load_parts_from_file(self):
        if not isfile(self.tracker_file):
            return
        with open(self.tracker_file, "r") as f:
            header_blob = json.loads(f.readline())
            self.download_url = header_blob["download_url"]
            start_time = header_blob["start_time"]  # for now we don't expire resumable downloads
            self.download_started = True
            for line in f:
                part_info = json.loads(line)
                part_id = part_info["part_number"]
                self._loaded_parts[part_id] = part_info

    def part_has_been_downloaded(self, part_number):
        if not self.open:
            return False
        if part_number not in self._loaded_parts:
            return False
        part_info = self._loaded_parts[part_number]
        part_path = part_info["part_filename"]
        return isfile(part_path)
    
    def get_part_info(self, part_number):
        if not self.open:
            return None
        return self._loaded_parts.get(part_number, None)
    
    def cleanup(self):
        if not self.open:
            return
        for part in self._loaded_parts.values():
            part_path = part["part_filename"]
            if isfile(part_path):
                os.remove(part_path)
        os.remove(self.tracker_file)
        self.open = False
    




import time
import json
import os
from os.path import basename, getsize, join, dirname, isfile, getctime
from pathlib import Path
from random import random
import requests

from geoseeq.knex import GeoseeqGeneralError
from geoseeq.constants import FIVE_MB
from geoseeq.utils import md5_checksum
from concurrent.futures import ThreadPoolExecutor, as_completed
from .utils import *
from geoseeq.file_system_cache import GEOSEEQ_CACHE_DIR
from .file_chunker import FileChunker
from .resumable_upload_tracker import ResumableUploadTracker


class ResultFileUpload:
    """Abstract class that handles upload methods for result files."""

    def _result_type(self, atomic=False):
        if self.is_sample_result:
            return "sample"
        if atomic:
            return "project"
        return "group"

    def _create_multipart_upload(self, filepath, file_size, optional_fields, atomic=False):
        optional_fields = optional_fields if optional_fields else {}
        optional_fields.update(
            {
                "md5_checksum": md5_checksum(filepath),
                "file_size_bytes": file_size,
            }
        )
        data = {
            "filename": basename(filepath),
            "optional_fields": optional_fields,
            "result_type": self._result_type(atomic),
        }
        url = f"/ar_fields/{self.uuid}/create_upload"
        if atomic:
            data["fieldname"] = self.name
            url = f"/ars/{self.parent.uuid}/create_atomic_upload"
        response = self.knex.post(url, json=data)
        return response
    
    def _prep_multipart_upload(self, filepath, file_size, chunk_size, optional_fields, atomic=False):
        n_parts = int(file_size / chunk_size) + 1
        response = self._create_multipart_upload(filepath, file_size, optional_fields, atomic=atomic)
        upload_id = response["upload_id"]
        data = {
            "parts": list(range(1, n_parts + 1)),
            "stance": "upload-multipart",
            "upload_id": upload_id,
            "result_type": self._result_type(atomic),
        }
        url = f"/ar_fields/{self.uuid}/create_upload_urls"
        if atomic:
            data["uuid"] = response["uuid"]
            data["fieldname"] = self.name
            url = f"ars/{self.parent.uuid}/create_atomic_upload_urls"
        response = self.knex.post(url, json=data)
        urls = response
        return upload_id, urls
    
    def _upload_one_part(self, file_chunker, url, num, max_retries, session=None, resumable_upload_tracker=None):
        if resumable_upload_tracker and resumable_upload_tracker.part_has_been_uploaded(num + 1):
            logger.info(f"Part {num + 1} has already been uploaded. Skipping.")
            return resumable_upload_tracker.get_part_info(num + 1)
        file_chunk = file_chunker.get_chunk(num)
        attempts = 0
        while attempts < max_retries:
            try:
                # url = url.replace("s3.wasabisys.com", "s3.us-east-1.wasabisys.com")
                logger.debug(f"Uploading part {num + 1} to {url}. Size: {len(file_chunk)} bytes.")
                if session:
                    http_response = session.put(url, data=file_chunk)
                else:
                    http_response = requests.put(url, data=file_chunk)
                http_response.raise_for_status()
                logger.debug(f"Upload for part {num + 1} succeeded.")
                break
            except (requests.exceptions.HTTPError, requests.exceptions.SSLError, requests.exceptions.ConnectionError) as e:
                attempts += 1
                logger.debug(
                    f"Upload for part {num + 1} failed. Attempt {attempts} of {max_retries}. Error: {e}"
                )
                if attempts >= max_retries:
                    raise e

                retry_time = min(8 ** attempts, 120)  # exponential backoff, max 120s
                retry_time *= 0.6 + (random() * 0.8)  # randomize to avoid thundering herd
                logger.debug(f"Retrying upload for part {num + 1} in {retry_time} seconds.")
                time.sleep(retry_time)
            
        etag = http_response.headers["ETag"].replace('"', "")
        blob = {"ETag": etag, "PartNumber": num + 1}
        if resumable_upload_tracker:
            # TODO technically not thread safe, but should be fine for now
            resumable_upload_tracker.add_part(blob)
        return blob
    
    def _finish_multipart_upload(self, upload_id, complete_parts, atomic=False):
        data = {
            "parts": complete_parts,
            "upload_id": upload_id,
            "result_type": self._result_type(atomic),
        }
        url = f"/ar_fields/{self.uuid}/complete_upload"
        if atomic:
            data["fieldname"] = self.name
            url = f"/ars/{self.parent.uuid}/complete_atomic_upload"
        response = self.knex.post(url, json=data, json_response=False)
        response.raise_for_status()

    def _upload_parts(self, file_chunker, urls, max_retries, session, progress_tracker, threads, resumable_upload_tracker=None):
        if threads == 1:
            logger.info(f"Uploading parts in series for {file_chunker.filepath}")
            complete_parts = []
            for num, url in enumerate(list(urls.values())):
                response_part = self._upload_one_part(file_chunker, url, num, max_retries, session, resumable_upload_tracker)
                complete_parts.append(response_part)
                if progress_tracker: progress_tracker.update(file_chunker.get_chunk_size(num))
                logger.info(f'Uploaded part {num + 1} of {len(urls)} for "{file_chunker.filepath}"')
            return complete_parts
        
        with ThreadPoolExecutor(max_workers=threads) as executor:
            logger.info(f"Uploading parts in parallel for {file_chunker.filepath} with {threads} threads.")
            futures = []
            for num, url in enumerate(list(urls.values())):
                future = executor.submit(
                    self._upload_one_part, file_chunker, url, num, max_retries, session, resumable_upload_tracker
                )
                futures.append(future)
            complete_parts = []
            for future in as_completed(futures):
                response_part = future.result()
                complete_parts.append(response_part)
                if progress_tracker: progress_tracker.update(file_chunker.get_chunk_size(response_part["PartNumber"] - 1))
                logger.info(
                    f'Uploaded part {response_part["PartNumber"]} of {len(urls)} for "{file_chunker.filepath}"'
                )
        complete_parts = sorted(complete_parts, key=lambda x: x["PartNumber"])
        return complete_parts

    def multipart_upload_file(
        self,
        filepath,
        file_size,
        optional_fields=None,
        chunk_size=None,
        max_retries=3,
        session=None,
        progress_tracker=None,
        threads=1,
        use_cache=True,
        use_atomic_upload=False,
    ):
        """Upload a file to S3 using the multipart upload process."""
        logger.info(f"Uploading {filepath} to S3 using multipart upload.")
        if not chunk_size:
            chunk_size = FIVE_MB
            if file_size >= 10 * FIVE_MB:
                chunk_size = 5 * FIVE_MB
        logger.debug(f"Using chunk size of {chunk_size} bytes.")
        resumable_upload_tracker = None
        if use_cache and file_size > 10 * FIVE_MB:  # only use resumable upload tracker for larger files
            upload_target_uuid = self.parent.uuid if use_atomic_upload else self.uuid
            resumable_upload_tracker = ResumableUploadTracker(filepath, chunk_size, upload_target_uuid)

        if resumable_upload_tracker and resumable_upload_tracker.upload_started:
            # a resumable upload for this file has already started
            resumable_upload_exists_and_is_valid = True
            upload_id, urls = resumable_upload_tracker.upload_id, resumable_upload_tracker.urls
            use_atomic_upload = resumable_upload_tracker.is_atomic_upload
            logger.info(f'Resuming upload for "{filepath}", upload_id: "{upload_id}"')
        else:
            upload_id, urls = self._prep_multipart_upload(filepath, file_size, chunk_size, optional_fields, atomic=use_atomic_upload)
            if resumable_upload_tracker:
                logger.info(f'Creating new resumable upload for "{filepath}", upload_id: "{upload_id}"')
                resumable_upload_tracker.start_upload(upload_id, urls, is_atomic_upload=use_atomic_upload)

        logger.info(f'Starting upload for "{filepath}"')
        complete_parts = []
        file_chunker = FileChunker(filepath, chunk_size)
        if file_chunker.file_size < 10 * FIVE_MB:
            file_chunker.load_all_chunks()
            logger.debug(f"Preloaded all chunks for {filepath}")
        else:
            logger.debug(f"Did not preload chunks for {filepath}")
        if progress_tracker: progress_tracker.set_num_chunks(file_chunker.file_size)
        complete_parts = self._upload_parts(
            file_chunker,
            urls,
            max_retries,
            session,
            progress_tracker,
            threads,
            resumable_upload_tracker=resumable_upload_tracker
        )
        self._finish_multipart_upload(upload_id, complete_parts, atomic=use_atomic_upload)
        logger.info(f'Finished Upload for "{filepath}"')
        if use_atomic_upload:
            # if this was an atomic upload then this result may not have existed on the server before
            self.get()
        return self

    def upload_file(self, filepath, multipart_thresh=FIVE_MB, overwrite=True, no_new_versions=False, **kwargs):
        if self.exists() and not overwrite:  
            raise GeoseeqGeneralError(f"Overwrite is set to False and file {self.uuid} already exists.")
        if not kwargs.get("use_atomic_upload", False):
            self.idem()
        else:
            self.parent.idem()
        if no_new_versions and self.has_downloadable_file():
            raise GeoseeqGeneralError(f"File {self} already has a downloadable file. Not uploading a new version.")
        resolved_path = Path(filepath).resolve()
        file_size = getsize(resolved_path)
        return self.multipart_upload_file(filepath, file_size, **kwargs)
    
    def upload_json(self, data, **kwargs):
        """Upload a file with the given data as JSON."""
        with NamedTemporaryFile("w", suffix='.json') as f:
            json.dump(data, f)
            f.flush()
            return self.upload_file(f.name, **kwargs)



import click
import json
from .shared_params import use_common_state, overwrite_option
from geoseeq import GeoseeqNotFoundError
from geoseeq.blob_constructors import (
    sample_result_file_from_uuid,
    project_result_file_from_uuid,
    sample_result_folder_from_uuid,
    project_result_folder_from_uuid,
)


@click.group('raw')
def cli_raw():
    """Low-level commands for interacting with the API."""
    pass


@cli_raw.command('get-file-data')
@use_common_state
@click.argument('file_ids', nargs=-1)
def cli_get_file_data(state, file_ids):
    """Print the raw stored data in a result file object."""
    knex = state.get_knex()
    for file_id in file_ids:
        file_id = file_id.split(':')[-1]
        try:
            result_file = sample_result_file_from_uuid(knex, file_id)
        except GeoseeqNotFoundError:
            result_file = project_result_file_from_uuid(knex, file_id)
        print(json.dumps(result_file.stored_data, indent=2), file=state.outfile)


@cli_raw.command('create-raw-file')
@use_common_state
@overwrite_option
@click.argument('folder_id')
@click.argument('result_filename')
@click.argument('filename', type=click.File('r'))
def cli_get_file_data(state, overwrite, folder_id, result_filename, filename):
    """Print the raw stored data in a result file object."""
    knex = state.get_knex()

    folder_id = folder_id.split(':')[-1]
    try:
        result_folder = sample_result_folder_from_uuid(knex, folder_id)
    except GeoseeqNotFoundError:
        result_folder = project_result_folder_from_uuid(knex, folder_id)
    blob = json.load(filename)
    result_file = result_folder.result_file(result_filename)
    if overwrite:
        result_file.idem()
        result_file.stored_data = blob
        result_file.save()
    else:
        result_file.create()
    click.echo(f'Created file {result_file.uuid}', file=state.outfile)
    





import urllib.request
import logging
import requests
import os
from os.path import basename, getsize, join, isfile, getmtime, dirname
from pathlib import Path
from tempfile import NamedTemporaryFile
from math import ceil

from geoseeq.utils import download_ftp
from geoseeq.constants import FIVE_MB
from hashlib import md5
from .resumable_download_tracker import ResumableDownloadTracker

logger = logging.getLogger("geoseeq_api")  # Same name as calling module

def url_to_id(url):
    url = url.split("?")[0]
    return md5(url.encode()).hexdigest()[:16]


def _download_head(url, filename, head=None, start=0, progress_tracker=None):
    headers = None
    if head and head > 0:
        headers = {"Range": f"bytes={start}-{head}"}
    response = requests.get(url, stream=True, headers=headers)
    response.raise_for_status()
    total_size_in_bytes = int(response.headers.get('content-length', 0))
    if progress_tracker: progress_tracker.set_num_chunks(total_size_in_bytes)
    if total_size_in_bytes > 10 * FIVE_MB:  # Use resumable download
        print("Using resumable download")
        return _download_resumable(response, filename, total_size_in_bytes, progress_tracker)
    else:
        block_size = FIVE_MB
        with open(filename, 'wb') as file:
            for data in response.iter_content(block_size):
                if progress_tracker: progress_tracker.update(len(data))
                file.write(data)
        return filename
    

def _download_resumable(response, filename, total_size_in_bytes, progress_tracker=None, chunk_size=5 * FIVE_MB, part_prefix=".gs_download_{}_{}."):
    target_id = url_to_id(response.url)
    tracker = ResumableDownloadTracker(chunk_size, target_id, filename)
    if not tracker.download_started: tracker.start_download(response.url)
    n_chunks = ceil(total_size_in_bytes / chunk_size)
    for i in range(n_chunks):
        bytes_start, bytes_end = i * chunk_size, min((i + 1) * chunk_size - 1, total_size_in_bytes - 1)
        if tracker.part_has_been_downloaded(i):
            logger.debug(f"Part {i} has already been downloaded.")
        else:
            logger.debug(f"Downloading part {i} of {n_chunks - 1}")
            part_filename = join(dirname(filename), part_prefix.format(i, n_chunks - 1) + basename(filename))
            _download_head(response.url, part_filename, head=bytes_end, start=bytes_start, progress_tracker=None)
            part_info = dict(part_number=i, start=bytes_start, end=bytes_end, part_filename=part_filename)
            tracker.add_part(part_info)
        if progress_tracker: progress_tracker.update(bytes_end - bytes_start + 1)
        
    # at this point all parts have been downloaded
    with open(filename, 'wb') as file:
        for i in range(n_chunks):
            part_info = tracker.get_part_info(i)
            part_filename = part_info["part_filename"]
            with open(part_filename, 'rb') as part_file:
                file.write(part_file.read())
    tracker.cleanup()
    return filename


def _download_generic(url, filename, head=None):
    urllib.request.urlretrieve(url, filename)
    return filename


def guess_download_kind(url):
    if 'azure' in url:
        return 'azure'
    elif 's3' in url:
        return 's3'
    elif 'ftp' in url:
        return 'ftp'
    else:
        return 'generic'


def download_url(url, kind='guess', filename=None, head=None, progress_tracker=None, target_uuid=None):
    """Return a local filepath to the downloaded file. Download the file."""
    if filename and isfile(filename):
        file_size = getsize(filename)
        if file_size > 0:
            logger.info(f"File already exists: {filename}. Not overwriting.")
            return filename
    if kind == 'guess':
        kind = guess_download_kind(url)
        logger.info(f"Guessed download kind: {kind} for {url}")
    logger.info(f"Downloading {kind} file to {filename}")
    if kind == 'generic':
        return _download_generic(url, filename, head=head)
    elif kind == 's3':
        return _download_head(url, filename, head=head, progress_tracker=progress_tracker)
    elif kind == 'azure':
        return _download_head(url, filename, head=head)
    elif kind == 'ftp':
        return download_ftp(url, filename, head=head)
    else:
        raise ValueError(f"Unknown download kind: {kind}")


class ResultFileDownload:
    """Abstract class that handles download methods for result files."""

    def get_download_url(self):
        """Return a URL that can be used to download the file for this result."""
        blob_type = self.stored_data.get("__type__", "").lower()
        if blob_type not in ["s3", "sra", "ftp", "azure"]:
            raise ValueError(f'Unknown URL type: "{blob_type}"')
        key = 'url' if 'url' in self.stored_data else 'uri'
        if blob_type in ["s3", "azure"]:
            try:
                url = self.stored_data["presigned_url"]
            except KeyError:
                url = self.stored_data[key]
            if url.startswith("s3://"):
                url = self.stored_data["endpoint_url"] + "/" + url[5:]
            return url
        else:
            return self.stored_data[key]
        
    
    def _download_flag_path(self, filename, flag_suffix='.gs_downloaded'):
        return filename + flag_suffix
        
    def download_needs_update(self, filename, flag_suffix='.gs_downloaded', slack=5):
        """Return True if the file needs to be downloaded, False otherwise.
        
        If either the file or the flag file does not exist, return True.
        If the flag file is older than `updated_at` in the result, return True.
        Otherwise, return False.
        """
        if isfile(filename) and isfile(self._download_flag_path(filename, flag_suffix)):
            if self.updated_at_timestamp - getmtime(self._download_flag_path(filename, flag_suffix)) > slack:
                return True
            return False
        return True

    def download(self, filename=None, flag_suffix='.gs_downloaded', cache=True, head=None, progress_tracker=None):
        """Return a local filepath to the file in this result. Download the file if necessary.
        
        When the file is downloaded, it is cached in the result object. Subsequent calls to download
        on this object will return the cached file unless cache=False is specified or the file is updated
        on the server.

        A flag file is created when the file download is complete. Subsequent calls to download
        will return the cached file if the flag file exists unless cache=False is specified.
        """
        if not filename and not self._cached_filename:
            self._temp_filename = True
            myfile = NamedTemporaryFile(delete=False)
            myfile.close()
            filename = myfile.name
        elif not filename and self._cached_filename:
            filename = self._cached_filename

        blob_type = self.stored_data.get("__type__", "").lower()
        needs_update = self.download_needs_update(filename, flag_suffix)
        if not needs_update:
            if cache and self._cached_filename:
                return self._cached_filename
            if cache and flag_suffix:
                # check if file and flag file exist, if so, return filename
                if isfile(filename) and isfile(self._download_flag_path(filename, flag_suffix)):
                    return filename

        url = self.get_download_url()
        filepath = download_url(
            url, blob_type, filename,
            head=head, progress_tracker=progress_tracker,
        )
        if cache and flag_suffix:
            # create flag file
            open(self._download_flag_path(filename, flag_suffix), 'a').close()
        if cache:
            self._cached_filename = filepath
        return filepath
